import datetime
import inspect
import typing

import mincepy


def pretty_type_string(obj_type: typing.Type) -> str:
    """Given an type will return a simple type string"""
    type_str = str(obj_type)
    if type_str.startswith('<class '):
        return type_str[8:-2]
    return type_str


def format_record(record: mincepy.DataRecord, user=False, ctime=False, version=False) -> list:
    hist = mincepy.get_historian()
    try:
        type_str = pretty_type_string(hist.get_helper(record.type_id).TYPE)
    except TypeError:
        type_str = str(record.type_id)

    data = []
    data.append(type_str)

    if user:
        data.append(record.get_extra(mincepy.ExtraKeys.USER))

    if version:
        data.append(record.version)

    if ctime:
        data.append(pretty_datetime(record.creation_time))

    try:
        hist.get_obj(record.obj_id)
        is_loaded = True
    except mincepy.NotFound:
        is_loaded = False

    data.append(record.obj_id)
    data.append('*' if is_loaded else '')

    return data


def obj_dict(obj):
    """Given an object return a dictionary that represents it"""
    repr_dict = {}
    for name in dir(obj):
        if not name.startswith('_'):
            try:
                value = getattr(obj, name)
                if not inspect.isroutine(value):
                    repr_dict[name] = value
            except Exception as exc:  # pylint: disable=broad-except
                repr_dict[name] = '{}: {}'.format(type(exc).__name__, exc)

    return repr_dict


def pretty_datetime(value) -> str:
    if isinstance(value, type):
        return pretty_type_string(value)
    if isinstance(value, datetime.datetime):
        if value.year == datetime.datetime.now().year:
            fmt = "%b %d %H:%M"
        else:
            fmt = "%b %d %Y"

        return value.strftime(fmt)

    return str(value)



import pytest

from mincepy.testing import Car

from pyos import pyos, nodes


def test_ls_basic():
    assert len(pyos.ls()) == 0

    car1 = Car()
    car1.save()

    assert len(pyos.ls()) == 1

    # Now save in a different directory
    pyos.cd('test')

    assert len(pyos.ls()) == 0

    car = Car()
    car.save()

    assert len(pyos.ls()) == 1

    pyos.cd('..')
    contents = pyos.ls()
    assert len(contents) == 2  # Now there is a directory and a file
    found = []
    for entry in contents:
        if isinstance(entry, nodes.DirectoryNode):
            assert entry.name == 'test/'
            found.append(True)
            continue

        if isinstance(entry, nodes.ObjectNode):
            assert entry.obj_id == car1.obj_id
            found.append(True)
            continue
    assert len(found) == 2


def test_ls_path():
    """Test that ls lists the contents of a folder when given a path"""
    car = Car()
    pyos.save(car, 'a/')

    assert len(pyos.ls()) == 1  # Should have the directory in home
    assert len(pyos.ls('a/')) == 1


def test_ls_dirs():
    subdirs = ['a/', 'b/', 'c/', 'd/']
    for subdir in subdirs:
        # Put a couple of cars in just to make it more realistic
        pyos.save(Car(), subdir)
        pyos.save(Car(), subdir)

    results = pyos.ls()
    assert len(results) == len(subdirs)
    for result in results:
        assert isinstance(result, nodes.DirectoryNode)
        idx = subdirs.index(result.name)
        assert idx != -1
        subdirs.pop(idx)
    assert not subdirs


def test_ls_minus_d():
    # Two cars at top level and two in the garage
    pyos.save(Car())
    pyos.save(Car())
    pyos.save(Car(), 'garage/')
    pyos.save(Car(), 'garage/')

    # The two cars, plus the directory
    assert len(pyos.ls()) == 3

    # Just the current directory
    results = pyos.ls(-pyos.d)
    assert len(results) == 1
    assert results[0].abspath == pyos.pwd()


def test_ls_inexistent():
    with pytest.raises(ValueError):
        assert not pyos.ls('inexistent')



__all__ = 'ValueOp', 'Flag', 'Options', 'separate_opts'


class ValueOp:
    """An operator that has a value"""

    def __init__(self, name: str, **kwargs):
        self._name = name
        self._kwargs = kwargs

    def __str__(self):
        return self._name

    def __eq__(self, other) -> bool:
        if not isinstance(other, ValueOp):
            return False

        if not self.name == other.name:
            return False

        try:
            return self.get() == other.get()
        except RuntimeError:
            return False

    def __neg__(self):
        """Enable the familiar command line -[op] notation, e.g. -l.
        This also allows for double negation i.e. --l"""
        return self

    def __call__(self, val):
        """pass in a value for this option"""
        if hasattr(self, '_val'):
            raise RuntimeError("Value already set!")
        option = ValueOp(self.name, **self._kwargs)
        setattr(option, '_val', val)
        return option

    @property
    def name(self):
        return self._name

    @property
    def default(self):
        try:
            return self._kwargs['default']
        except KeyError:
            raise AttributeError('default')

    def has_default(self):
        return 'default' in self._kwargs

    def get(self, *args):
        if len(args) > 1:
            raise ValueError("Takes only one argument, a default value")

        try:
            return getattr(self, '_val')
        except AttributeError:
            pass

        # Look for a default
        if args:
            return args[0]

        try:
            return self._kwargs['default']
        except KeyError:
            pass

        raise ValueError("This option has no set value or default")


class Flag(ValueOp):

    def __init__(self, name):
        super().__init__(name)
        self._val = True

    def __call__(self, val):
        raise RuntimeError("A flag does not take values")

    def get(self, *args):
        if args:
            raise ValueError("A flag cannot be passed a default value, it is always True")
        return True


class Options:
    """Container storing options, using this with seperate_opts makes writing tools much easier"""

    def __init__(self):
        self._opts = {}

    def __contains__(self, item):
        if not isinstance(item, ValueOp):
            return False

        return item.name in self._opts

    def add(self, opt: ValueOp):
        self._opts[opt.name] = opt

    def pop(self, opt: ValueOp, *default):
        if not isinstance(opt, ValueOp):
            raise TypeError("Unsupported option type '{}'".format(type(opt)))
        if len(default) > 1:
            raise ValueError("Can only pass at most one default")

        name = opt.name

        try:
            found = self._opts.pop(name)
        except KeyError:
            if isinstance(opt, Flag):
                return False
            if default:
                return default[0]
            if opt.has_default():
                return opt.default

            raise

        try:
            return found.get(*default)
        except ValueError:
            raise KeyError("Option '{}' does not have a value".format(name))


def separate_opts(*args) -> [Options, list]:
    """Separate out the options from the other arguments"""
    opts = Options()
    rest = []
    for arg in args:
        if isinstance(arg, ValueOp):
            opts.add(arg)
        else:
            rest.append(arg)
    return opts, rest



"""Tests for the change directory command"""

from pyos import pyos


def test_cd_simple():
    start = pyos.pwd()
    assert isinstance(start, pyos.PyosPath)

    # Try relative directory changes
    a = 'a/'
    pyos.cd(a)
    assert pyos.pwd() == start / a

    b = 'b/'
    pyos.cd(b)
    assert pyos.pwd() == start / a / b

    # Try absolute directory change
    pyos.cd(start / a)
    assert pyos.pwd() == start / a

    # Try .. change
    pyos.cd('..')
    assert pyos.pwd() == start



from mincepy.testing import Car

from pyos import pyos


def test_load_single():
    """Test loading a single objects just returns that object and not a list"""
    car = Car()
    car_id = car.save()
    loaded = pyos.load(car_id)
    assert loaded is car


def test_load_from_path():
    car = Car()
    pyos.save(car, 'my_car')

    # Try with string first
    assert pyos.load('my_car') is car

    # Try with path
    assert pyos.load(pyos.PyosPath('my_car')) is car



from mincepy.testing import Car

from pyos import pyos


def test_oid_simple():
    car = Car()
    car.save()

    assert pyos.oid(car) == car.obj_id

    # Save some more and see if the order is preserved
    cars = []
    for _ in range(10):
        cars.append(Car())

    pyos.save(*cars)
    assert pyos.oid(*cars) == [car.obj_id for car in cars]



"""Module containig standard option names"""

from . import opts

__all__ = 'l', 'L', 'd', 'mindepth', 'maxdepth'

# pylint: disable=invalid-name

# region Options
l = opts.Flag('l')
d = opts.Flag('d')
L = opts.ValueOp('L')
maxdepth = opts.ValueOp('maxdepth')
mindepth = opts.ValueOp('mindepth')

# endregion



import pprint
import typing

import mincepy

from . import dirs
from . import fmt
from . import lib
from . import nodes
from .dirs import PyosPath
from . import opts
from . import sopts
from . import utils
from .sopts import *  # pylint: disable=wildcard-import

# pylint: disable=invalid-name

__all__ = ('pwd', 'cd', 'ls', 'load', 'save', 'cat', 'locate', 'mv', 'meta', 'rm', 'find',
           'history', 'tree') + sopts.__all__


def pwd():
    """Return the current working directory"""
    return dirs.cwd()


def cd(path: dirs.PathSpec):
    path = PyosPath(path)
    if path.is_file():
        # Assume they just left out the slash
        path = path.to_dir()

    dirs.cd(path)


def ls(*args, _type: typing.Type = None) -> nodes.ResultsNode:
    """List the contents of a directory

    :type: restrict listing to a particular type
    """
    options, rest = opts.separate_opts(*args)
    parsed = utils.parse_args(*rest)

    results = nodes.ResultsNode()
    if rest:
        for entry in parsed:
            if isinstance(entry, Exception):
                raise entry

            results.append(nodes.to_node(entry))
    else:
        results.append(nodes.to_node(dirs.cwd()))

    if not options.pop(sopts.d):
        for entry in results:
            if isinstance(entry, nodes.DirectoryNode):
                entry.expand()

        if len(results) == 1 and isinstance(results[0], nodes.DirectoryNode):
            sole_dir = results[0]
            new_results = nodes.ResultsNode(sole_dir.name)
            for result in tuple(sole_dir.children):
                result.parent = new_results

            results = new_results

    if options.pop(sopts.l):
        results.show('loaded', 'type', 'creator', 'version', 'mtime', 'name', mode=nodes.TABLE_VIEW)
    else:
        results.show(mode=nodes.LIST_VIEW)

    return results


def load(*obj_or_ids) -> typing.Union[typing.Iterable[typing.Any], typing.Any]:
    """Load one or more objects"""
    _options, rest = opts.separate_opts(*obj_or_ids)
    to_load = ls(-d, *rest)

    loaded = []
    for node in to_load:
        try:
            loaded.append(mincepy.load(node.obj_id))
        except Exception as exc:  # pylint: disable=broad-except
            loaded.append(exc)

    if len(loaded) == 1:
        return loaded[0]

    return loaded


def save(*objs):
    """Save one or more objects"""

    if len(objs) > 1 and isinstance(objs[-1], (str, dirs.PyosPath)):
        # Extract the destination
        dest = PyosPath(objs[-1]).resolve()
        objs = objs[:-1]

        if len(objs) > 1 and dest.is_file():
            # Automatically convert to directory if there are many objects as they can't save
            # more than one with the same filename in the same folder!
            dest = dest.to_dir()

        return lib.save(objs, [dest] * len(objs))

    return lib.save(objs)


def cat(*obj_or_ids):
    """Print the contents of one or more objects"""
    _options, rest = opts.separate_opts(*obj_or_ids)
    to_cat = load(*rest)

    for obj in to_cat:
        if isinstance(obj, Exception):
            print(obj)
        else:
            if isinstance(obj, mincepy.File):
                print(obj.read_text())
            else:
                pprint.pprint(fmt.obj_dict(obj))


def locate(*obj_or_ids) -> nodes.ResultsNode:
    """Locate the directory of or more objects"""
    _options, rest = opts.separate_opts(*obj_or_ids)
    to_locate = ls(-d, *rest)
    to_locate.show('abspath', mode=nodes.TABLE_VIEW)

    return to_locate


def mv(*args):  # pylint: disable=invalid-name
    """Take one or more files or directories with the final parameter being interpreted as
     destination"""
    _options, rest = opts.separate_opts(*args)
    assert len(rest) <= 2, "mv: missing destination"
    dest = dirs.PyosPath(rest.pop())
    if len(rest) > 1:
        # If there is more than one thing to move then we assume that dest is a directory
        dest = dest.to_dir()
    dest = dest.resolve()
    to_move = ls(*rest[:-1])
    to_move.move(dest)


def rm(*obj_or_ids):
    _options, rest = opts.separate_opts(*obj_or_ids)
    to_delete = ls(-d, *rest)
    for node in to_delete:
        node.delete()


def meta(*obj_or_ids, **updates):
    """Get or update the metadata on one or more objects"""
    _options, rest = opts.separate_opts(*obj_or_ids)
    to_update = ls(-d, *rest)
    obj_ids = []
    for node in to_update:
        if isinstance(node, nodes.ObjectNode):
            obj_ids.append(node.obj_id)
        else:
            print("Can't set metadata on '{}'".format(node))

    if updates:
        # In 'setting' mode
        lib.update_meta(*obj_ids, meta=updates)
    else:
        # In 'getting' mode
        metas = lib.get_meta(*obj_ids)
        if len(metas) == 1:
            return metas[0]

        return metas

    return None


def find(*args, **meta_filter):
    return lib.find(*args, **meta_filter)


def history(obj):
    hist = mincepy.get_historian()
    for entry in hist.history(obj):
        print(entry.ref)
        cat(entry.obj)
        print()


def tree(*paths):
    """Get a tree representation of the given paths"""
    options, rest = opts.separate_opts(*paths)
    to_tree = ls(-d, *rest)
    level = options.pop(sopts.L, -1)
    # Fully expand all directories
    for dir_node in to_tree.directories:
        dir_node.expand(level)
    to_tree.show(mode=nodes.TREE_VIEW)
    return to_tree


def oid(*args):
    """Get the object id for one or more live objects"""
    _options, rest = opts.separate_opts(*args)
    hist = mincepy.get_historian()

    oids = []
    for obj in rest:
        try:
            oids.append(hist.get_obj_id(obj))
        except mincepy.NotFound:
            oids.append(None)

    if len(oids) == 1:
        return oids[0]

    return oids



from pyos import pyos
from pyos import dirs


def test_reaching_root():
    """Assert the correct behaviour when traversing up until reaching root and beyond"""
    cwd = dirs.cwd()
    for _ in range(len(cwd.parts) - 1):
        pyos.cd('..')

    # Should be at root
    assert dirs.cwd() == dirs.PyosPath('/')

    # Now trying going up again
    pyos.cd('..')
    # Should still be at root
    assert dirs.cwd() == dirs.PyosPath('/')


def test_saving_path():
    path = dirs.PyosPath('/some/random/path/')
    path_str = str(path)
    obj_id = pyos.save(path)
    del path

    loaded = pyos.load(obj_id)
    assert str(loaded) == path_str



from abc import ABCMeta
from collections.abc import Sequence
import typing

import anytree
import columnize
import tabulate

import mincepy

from . import constants
from . import fmt
from . import dirs
from . import queries


class BaseNode(Sequence, anytree.NodeMixin, metaclass=ABCMeta):

    def __init__(self, name: str, parent=None):
        super().__init__()
        self._name = name
        self.parent = parent
        self._hist = mincepy.get_historian()

    def __getitem__(self, item):
        return self.children.__getitem__(item)

    def __len__(self):
        return self.children.__len__()

    @property
    def name(self):
        return self._name

    def delete(self):
        """Delete this node and any descendents"""
        for child in self.children:
            child.delete()

    def move(self, where: dirs.PyosPath):
        """Move this object (with any descendents) to the given path"""
        for child in self.children:
            child.move(where)


class PyosNode(BaseNode):
    """Base node for representing an object in the virtual filesystem"""

    def __init__(self, abspath: dirs.PyosPath, parent=None):
        super().__init__(abspath.name, parent)
        self._abspath = abspath

    @property
    def abspath(self) -> dirs.PyosPath:
        return self._abspath


class DirectoryNode(PyosNode):

    def __init__(self, pathname: dirs.PyosPath, parent=None):
        assert pathname.is_dir(), "Must supply a directory path"
        super().__init__(pathname.resolve(), parent)

    def __repr__(self):
        rep = []
        for pre, _, node in anytree.RenderTree(self):
            rep.append("%s%s" % (pre, node.name))
        return "\n".join(rep)

    def __contains__(self, item):
        if isinstance(item, dirs.PyosPath):
            path = item.resolve()
            for child in self.children:
                if path == child.abspath:
                    return True

            return False

        if self._hist.is_obj_id(item):
            for node in self.objects:
                if item == node.obj_id:
                    return True
            return False

        return False

    @property
    def directories(self):
        return filter(lambda node: isinstance(node, DirectoryNode), self.children)

    @property
    def objects(self):
        return filter(lambda node: isinstance(node, ObjectNode), self.children)

    def _invalidate_cache(self):
        self.children = []

    def expand(self, depth=1):
        """Populate the children with what is currently in the database"""
        self.children = []
        metas = self._hist.meta.find(queries.subdirs(str(self._abspath), 0, -1))

        # Get directories and object ids
        dirstring = str(self._abspath)
        child_expand_depth = depth - 1
        directories_added = set()
        for meta in metas:
            directory = meta[constants.DIR_KEY]

            if directory == dirstring:
                # This is an object _in_ this directory
                ObjectNode(meta['obj_id'], meta=meta, parent=self)
            else:
                # This is an object that resides at some subdirectory so we know
                # that there must exist a subdirectory
                if directory not in directories_added:
                    directories_added.add(directory)

                    obj_dir = dirs.PyosPath(directory)
                    path = (self._abspath / obj_dir.parts[len(self._abspath.parts)]).to_dir()

                    dir_node = DirectoryNode(path, parent=self)
                    if abs(child_expand_depth) > 0:
                        dir_node.expand(child_expand_depth)

    def delete(self):
        metas = self._hist.meta.find(queries.subdirs(str(self._abspath), 0, -1))
        objs = self._hist.load(meta['obj_id'] for meta in metas)
        for obj in objs:
            self._hist.delete(obj)
        self._invalidate_cache()

    def move(self, where: dirs.PyosPath):
        assert where.is_dir(), "Can't move a directory to a file"
        where = where.resolve()
        self.expand(1)
        for child in self.children:
            child.move((where / self.name).to_dir())


class ObjectNode(PyosNode):

    @classmethod
    def from_path(cls, path: dirs.PyosPath):
        assert path.is_file()
        path = path.resolve()
        # Let's find the object id
        hist = mincepy.get_historian()
        try:
            obj_id = hist._ensure_obj_id(path.name)
            meta = None
        except mincepy.NotFound:
            results = tuple(
                hist.meta.find(
                    queries.and_(queries.dirmatch(path.parent), {constants.NAME_KEY: path.name})))
            if not results:
                raise ValueError("Object not found: {}".format(path))

            meta = results[0]
            obj_id = meta['obj_id']

        return ObjectNode(obj_id, meta=meta)

    def __init__(self, obj_id, record: mincepy.DataRecord = None, meta=None, parent=None):
        self._hist = mincepy.get_historian()
        self._obj_id = obj_id

        # Set up the record
        if record is None:
            self._record = tuple(self._hist.archive.find(obj_id))[0]  # type: mincepy.DataRecord
        else:
            self._record = record
        assert self._record.obj_id == obj_id

        # Set up the meta
        self._meta = meta
        if self._meta is None:
            try:
                self._meta = self._hist.meta.get(obj_id)
            except mincepy.NotFound:
                pass
        else:
            assert self._meta['obj_id'] == obj_id

        # Set up the abspath
        self._abspath = dirs.get_abspath(obj_id, self._meta)

        super().__init__(self._abspath, parent)

    def __contains__(self, item):
        return item == self.obj_id

    @property
    def loaded(self):
        try:
            self._hist.get_obj(self._obj_id)
            return True
        except mincepy.NotFound:
            return False

    @property
    def abspath(self) -> dirs.PyosPath:
        return self._abspath

    @property
    def obj_id(self):
        return self._obj_id

    @property
    def type(self):
        try:
            return self._hist.get_obj_type(self._record.type_id)
        except TypeError:
            return self._record.type_id

    @property
    def ctime(self):
        return self._record.creation_time

    @property
    def version(self):
        return self._record.version

    @property
    def mtime(self):
        return self._record.snapshot_time

    @property
    def creator(self):
        return self._record.get_extra(mincepy.ExtraKeys.CREATED_BY)

    @property
    def meta(self) -> dict:
        return self._meta

    def delete(self):
        self._hist.delete(self._hist.load_one(self._obj_id))

    def move(self, where: dirs.PyosPath):
        meta_update = dirs.path_to_meta_dict(where)
        self._hist.meta.update(self._obj_id, meta_update)


LIST_VIEW = 'list'
TREE_VIEW = 'tree'
TABLE_VIEW = 'table'


class ResultsNode(BaseNode):
    VIEW_PROPERTIES = ('loaded', 'type', 'creator', 'version', 'ctime', 'mtime', 'name', 'str')

    def __init__(self, name='results', parent=None):
        super().__init__(name, parent)
        self._view_mode = TABLE_VIEW
        self._show = {'name'}

    def __contains__(self, item):
        if isinstance(item, dirs.PyosPath):
            path = item.resolve()
            for child in self.children:
                if path == child.abspath:
                    return True

            return False

        if self._hist.is_obj_id(item):
            for node in self.objects:
                if item == node.obj_id:
                    return True
            return False

        return False

    def __repr__(self):
        if self._view_mode == TREE_VIEW:
            rep = []
            for child in self.directories:
                for pre, _, node in anytree.RenderTree(child):
                    rep.append("%s%s" % (pre, node.name))
            for child in self.objects:
                rep.append(str(child))
            return "\n".join(rep)

        if self._view_mode == TABLE_VIEW:
            rep = []

            if self._deeply_nested():
                # Do the objects first, like linux's 'ls'
                table = self._get_table(self.objects)
                rep.append(tabulate.tabulate(table, tablefmt='plain'))

                for directory in self.directories:
                    dir_repr = []
                    dir_repr.append("{}:".format(directory.name))
                    table = self._get_table(directory)
                    dir_repr.append(tabulate.tabulate(table, tablefmt='plain'))
                    rep.append("\n".join(dir_repr))
            else:
                my_repr = []
                table = self._get_table(self.directories)
                table.extend(self._get_table(self.objects))
                my_repr.append(tabulate.tabulate(table, tablefmt='plain'))
                rep.append("\n".join(my_repr))

            return "\n\n".join(rep)

        if self._view_mode == LIST_VIEW:

            repr_list = []
            for child in self:
                repr_list.append("-".join(self._get_row(child)))

            return columnize.columnize(repr_list, displaywidth=100)

        return super().__repr__()

    @property
    def directories(self):
        return filter(lambda node: isinstance(node, DirectoryNode), self.children)

    @property
    def objects(self):
        return filter(lambda node: isinstance(node, ObjectNode), self.children)

    def append(self, node: PyosNode, display_name: str = None):
        node.parent = self
        display_name = display_name or node.name
        node.display_name = display_name

    def show(self, *properties, mode: str = None):
        if mode is not None:
            self._view_mode = mode
        if properties:
            self._show = set(properties)

    def _get_row(self, child) -> typing.Sequence[str]:
        # pylint: disable=too-many-branches
        empty = ''
        row = []

        if 'loaded' in self._show:
            try:
                row.append('*' if child.loaded else '')
            except AttributeError:
                row.append(empty)

        if 'type' in self._show:
            try:
                row.append(fmt.pretty_type_string(child.type))
            except AttributeError:
                row.append('directory')

        if 'creator' in self._show:
            row.append(getattr(child, 'creator', empty))

        if 'version' in self._show:
            row.append(getattr(child, 'version', empty))

        if 'ctime' in self._show:
            try:
                row.append(fmt.pretty_datetime(child.ctime))
            except AttributeError:
                row.append(empty)

        if 'mtime' in self._show:
            try:
                row.append(fmt.pretty_datetime(child.mtime))
            except AttributeError:
                row.append(empty)

        if 'name' in self._show:
            row.append(getattr(child, 'name', empty))

        if 'str' in self._show:
            row.append(str(child))

        if 'abspath' in self._show:
            row.append(str(getattr(child, 'abspath', empty)))

        return row

    def _get_table(self, entry):
        table = []

        for child in entry:
            table.append(self._get_row(child))

        return table

    def _deeply_nested(self) -> bool:
        for directory in self.directories:
            if len(directory) > 0:
                return True

        return False


def to_node(entry) -> PyosNode:
    """Get the node for a given object.  This can be either:
    1. An object id -> ObjectNode
    2. A directory path -> DirectoryNode
    3. An object path -> ObjectNode
    """
    hist = mincepy.get_historian()
    if hist.is_obj_id(entry):
        return ObjectNode(entry)
    if isinstance(entry, dirs.PyosPath):
        if entry.is_dir():
            return DirectoryNode(entry)

        return ObjectNode.from_path(entry)

    raise ValueError("Unknown entry type: {}".format(entry))



"""Module that deals with directories and paths"""
import contextlib
import getpass
import os
import typing
import uuid

import mincepy

from .constants import DIR_KEY, NAME_KEY

__all__ = 'PyosPath', 'working_path'

_DIRECTORY = None  # type: typing.Optional[PyosPath]


class PyosPath(mincepy.BaseSavableObject):
    """A path in Pyos.  Where possible the convention follows that of a PurePosixPath in pathlib.
    The one major exception is that folders are represented with an explicit trailing '/' and
    anything else is a file.

    This is a 'pure' path in the pathlib sens in that it does not interact with the database at all
    """
    ATTRS = ('_path',)
    TYPE_ID = uuid.UUID('5eac541e-848c-43aa-818d-50cf8a2b8507')

    def __init__(self, path='./'):
        super().__init__()
        if isinstance(path, PyosPath):
            self._path = path._path
        elif isinstance(path, str):
            if path.endswith('.'):
                path = path + '/'
            self._path = path

    @property
    def parts(self) -> typing.Tuple[str]:
        parts = list(self._path.split('/'))
        if self.is_dir():
            parts.pop()
            parts[-1] = parts[-1] + '/'
            if self.is_absolute():
                parts[0] = '/'

        return tuple(parts)

    @property
    def parent(self):
        return PyosPath("/".join(self.parts[:-1]) + '/')

    @property
    def name(self):
        return self.parts[-1]

    @property
    def root(self) -> str:
        return "/"

    def __hash__(self):
        return self._path.__hash__()

    def __eq__(self, other):
        if not isinstance(other, PyosPath):
            return False

        return self._path == other._path

    def __repr__(self):
        return "PyosPath('{}')".format(self._path)

    def __str__(self):
        return self._path

    def __truediv__(self, other):
        if not isinstance(other, PyosPath):
            if isinstance(other, str):
                other = PyosPath(other)
            else:
                raise TypeError("Cannot join a path with a '{}'".format(type(other)))

        if self.is_file():
            raise ValueError("Can't join a file with another path")

        if other.is_absolute():
            return other

        return PyosPath(str(self) + str(other))

    def is_file(self) -> bool:
        return not self.is_dir()

    def is_dir(self) -> bool:
        return self._path.endswith('/')

    def is_absolute(self) -> bool:
        return self._path.startswith('/')

    def to_dir(self):
        """If this path is a file then a directory with the same name (and path) will be returned.
        Otherwise this path will be returned"""
        if self.is_dir():
            return self

        return PyosPath(str(self) + '/')

    def to_file(self):
        """If this path is a directory then a file with the same name (and path) will be returned.
        Otherwise this path will be returned"""
        if self.is_file():
            return self

        return PyosPath(str(self)[:-1])

    def resolve(self):
        """Make the path absolute eliminating any . and .. that occur in the path"""
        if not self.is_absolute():
            to_norm = cwd() / self
        else:
            to_norm = self

        # Now resolve any . and ..
        normed = os.path.normpath(str(to_norm))
        path = PyosPath(normed)

        if self.is_dir():
            # Re-add back the final slash
            return path.to_dir()

        return path

    def joinpath(self, *other):
        joined = self
        for entry in other:
            joined = joined / entry
        return joined


# The types that can be used to specify a path in PyOs
PathSpec = typing.Union[PyosPath, str]


def init():
    cd('/{}/'.format(getpass.getuser()))


def reset():
    global _DIRECTORY  # pylint: disable=global-statement
    mincepy.get_historian().meta.sticky.pop(DIR_KEY, None)
    _DIRECTORY = None


def cwd() -> PyosPath:
    """Get the current working directory"""
    global _DIRECTORY  # pylint: disable=global-statement
    return _DIRECTORY


def cd(directory: PathSpec) -> PyosPath:  # pylint: disable=invalid-name
    global _DIRECTORY  # pylint: disable=global-statement
    new_dir = PyosPath(directory).resolve()
    assert new_dir.is_dir(), "Cannot change directory to a file path"

    mincepy.get_historian().meta.sticky[DIR_KEY] = str(new_dir)
    _DIRECTORY = new_dir
    return _DIRECTORY


def get_obj_id(objname: str, directory: PathSpec = None):
    directory = (directory or cwd()).resolve()
    hist = mincepy.get_historian()

    metas = tuple(hist.meta.find({DIR_KEY: str(directory), NAME_KEY: objname}))
    if metas:
        assert len(metas) == 1, "Uniqueness constraint violated"
        return metas[0]['obj_id']

    return None


def get_abspath(obj_id, meta: dict):
    assert obj_id, "Must provide a valid obj id"
    meta = meta or {}
    directory = PyosPath(meta.get(DIR_KEY, '/'))
    name = meta.get(NAME_KEY, str(obj_id))
    return directory / name


# class Directory:
#
#     def __init__(self, name: PathSpec = PyosPath('/')):
#         self.__path = name.resolve()
#
#     def __str__(self):
#         return str(self.__path)
#
#     def __repr__(self):
#         return "Directory('{}')".format(self.__path)
#
#     def __dir__(self) -> typing.Iterable[str]:
#         """Get the contents of this directory as attributes.  We make the following mappings which
#         also mean that the attributes aren't unambigious but this can be resolved using the dir[]
#         notation.
#
#         The attributes are listed in the following priority:
#         1. Any objects in the directory are listed as dir._{obj_id}
#         2. Any objects with names are listed as dir.{obj_name} _unless_ there is a subdirectory
#            with the name {obj_name} in which case it is listed as dir.{obj_name}_{obj_id}
#         3. Subdirectories are listed as dir.{dir_name}
#
#         This has the following consequences (amongst others):
#         * Directories that have the name {obj_name}_{obj_id} will be hidden by point 2.
#         * Directories with the name _{obj_id} will be hidden by point 1.
#         """
#         objects, subdirs = get_contents(self.__path)
#
#         # Priority 3.
#         contents = subdirs
#
#         # Priority 2. Deduplicate names clashing with folders
#         for obj_id, name in objects.items():
#             if name in contents:
#                 contents.add("{}_{}".format(name, obj_id))
#             elif name:
#                 contents.add(name)
#
#         # Priority 1. Obj ids
#         contents.update("_{}".format(obj_id) for obj_id in objects)
#
#         return contents
#
#     def __getattr__(self, item: str):
#         """Uses priorities from __dir__ to get attribute from this directory"""
#         hist = mincepy.get_historian()
#
#         if not item:
#             raise AttributeError(item)
#
#         # Priority 1. Obj id
#         if item[0] == '_':
#             try:
#                 obj_id = hist._ensure_obj_id(item[1:])
#             except mincepy.NotFound:
#                 pass
#             else:
#                 return mincepy.load(obj_id)
#
#         # Priority 2. Objname
#         objects, subdirs = get_contents(self.__path)
#         # First check for the overwrite case
#         parts = item.split("_")
#         if len(parts) > 1:
#             hist = mincepy.get_historian()
#             try:
#                 obj_id = hist._ensure_obj_id(parts[-1])
#             except mincepy.NotFound:
#                 pass
#             else:
#                 if "_".join(parts[:-1]) in objects.values():
#                     return mincepy.load(obj_id)
#
#         # Check the objects for this name
#         for obj_id, name in objects.items():
#             if item == name:
#                 return mincepy.load(obj_id)
#
#         # Priority 3. Subdir
#         if item in subdirs:
#             return Directory(self.__path / item)
#
#         raise AttributeError(item)
#
#
# class CWD(Directory):
#
#     def __init__(self):
#         super(CWD, self).__init__()
#         self.__path = property(cwd)


def path_to_meta_dict(path: PathSpec):
    if path is None:
        return {}

    path = path.resolve()
    meta = {}
    if path.is_dir():
        meta[DIR_KEY] = str(path)
    else:
        meta[NAME_KEY] = path.name
        meta[DIR_KEY] = str(path.parent)

    return meta


@contextlib.contextmanager
def working_path(path: PyosPath):
    orig = cwd()
    cd(path)
    try:
        yield path
    finally:
        cd(orig)



from mincepy.testing import Car

from pyos import pyos


def test_save():
    car = Car()
    obj_id = pyos.save(car)
    assert obj_id == car.obj_id


def test_save_with_name():
    car = Car()
    pyos.save(car, 'my_car')
    results = pyos.ls('my_car')
    assert len(results) == 1
    assert results[0].name == 'my_car'


def test_save_to_dir():
    for _ in range(10):
        pyos.save(Car(), 'test/')

    results = pyos.ls('test/')
    assert len(results) == 10

    # Now check that save will promote a file to a directory
    pyos.save(Car(), 'test')
    assert len(pyos.ls('test/')) == 10



from mincepy.testing import Car

from pyos import pyos


def test_mv_basic():
    car = Car()
    car.save()

    # Move to test subdirectory
    pyos.mv(car, 'test/')

    contents = pyos.ls('test/')
    assert len(contents) == 1
    assert contents[0].obj_id == car.obj_id

    # Now move test into a subfolder
    pyos.mv('test/', 'sub/')
    contents = pyos.ls()
    assert len(contents) == 1
    assert contents[0].name == 'sub/'

    contents = pyos.ls('sub/')
    assert len(contents) == 1
    assert contents[0].name == 'test/'



from mincepy.testing import Person

import pyos


def test_working_path():
    home = pyos.pyos.pwd()
    address_book = pyos.PyosPath('address_book/').resolve()
    with pyos.working_path(address_book):
        person_id = Person('martin', 34).save()
        assert pyos.pyos.pwd() == home / address_book

    contents = pyos.pyos.ls()
    assert len(contents) == 1
    assert contents[0].abspath == address_book

    contents = pyos.pyos.ls(address_book)
    assert len(contents) == 1
    assert contents[0].obj_id == person_id



"""Module with convenience functions for building queries"""
from typing import Iterable

from .constants import DIR_KEY


def or_(*conditions):
    if len(conditions) == 1:
        return conditions[0]

    return {'$or': list(conditions)}


def and_(*conditions):
    return {'$and': list(conditions)}


def unset_(*keys: Iterable[str]):
    return {'$unset': {key: '' for key in keys}}


def subdirs(root: str, start_depth=1, end_depth=1) -> dict:
    """Get a query string that will look for subdirectories of root optionally specifying the
    start and end depths
    """
    if end_depth == -1:
        end_depth = ''  # This will cause the regex to allow any number of repetitions
    regex = ('^{}([^/]+/){{{},{}}}$'.format(root, start_depth, end_depth))
    return {DIR_KEY: {'$regex': regex}}


def dirmatch(directory: str) -> dict:
    """Get the query dictionary to search in a particular directory"""
    query = {DIR_KEY: str(directory)}
    if directory == "/":
        # Special case for root: all objects that have no DIR_KEY are by default
        # considered to be in the root
        query = or_(query, {DIR_KEY: {'$exists': False}})
    return query



from mincepy.testing import Car

from pyos import pyos


def fill_with_cars(subdirs: list):
    cwd = pyos.pwd()
    for subdir in subdirs:
        pyos.cd(subdir)
        car = Car()
        car.save(with_meta={'target': True, 'mydir': subdir})
    # Now change back to the original directory
    pyos.cd(cwd)


def test_simple_find():
    # Save a car
    car = Car()
    car.save(with_meta={'group': 'cars'})

    # Look for it
    results = pyos.find(group='cars')
    assert len(results) == 1
    assert results[0].obj_id == car.obj_id

    # Add another car to the group
    car2 = Car()
    car2.save(with_meta={'group': 'cars'})

    # Look for them
    results = pyos.find(group='cars')
    assert len(results) == 2
    assert car.obj_id in results
    assert car2.obj_id in results


def test_subdirs_find():
    subdirs = ['./', 'a/', 'b/', 'c/', 'd/']
    fill_with_cars(subdirs)
    num_subdirs = len(subdirs)

    results = pyos.find(target=True)
    assert len(results) == num_subdirs

    # Check mindepth
    for idx, subdir in enumerate(subdirs):
        found = pyos.find(pyos.mindepth(idx))
        assert len(found) == num_subdirs - idx
        dirs = {pyos.meta(node)['mydir'] for node in found}
        for check_dir in subdirs[idx:]:
            assert check_dir in dirs

    # Now check maxdepth
    for idx, subdir in enumerate(subdirs):
        # Note, have to use +1 on indexes here because of the 'exclusive' range notations, etc
        found = pyos.find(pyos.maxdepth(idx))
        assert len(found) == idx + 1
        dirs = {pyos.meta(node)['mydir'] for node in found}
        for check_dir in subdirs[:idx + 1]:
            assert check_dir in dirs

    # Now check combinations of mindepth and maxdepth

    for min_idx in range(len(subdirs)):
        for max_idx in range(min_idx, len(subdirs)):
            found = pyos.find(pyos.mindepth(min_idx), pyos.maxdepth(max_idx))
            assert len(found) == max_idx - min_idx + 1
            dirs = {pyos.meta(node)['mydir'] for node in found}

            for check_dir in subdirs[min_idx:max_idx + 1]:
                assert check_dir in dirs


def test_find_starting_point():
    """Test that find respects the passed starting points"""
    subdirs = ['./', 'a/', 'b/', 'c/', 'd/']
    fill_with_cars(subdirs)
    num_subdirs = len(subdirs)

    for idx, subdir in enumerate(subdirs):
        start_point = "/".join(subdirs[:idx + 1])
        found = pyos.find(start_point)
        assert len(found) == num_subdirs - idx
        dirs = {pyos.meta(meta_dict)['mydir'] for meta_dict in found}

        for check_dir in subdirs[idx:]:
            assert check_dir in dirs



import typing

import mincepy

from . import dirs
from . import nodes


def parse_args(*args) -> typing.Sequence:
    parsed = []
    hist = mincepy.get_historian()
    for arg in args:
        if arg is None:
            parsed.append(None)
        elif isinstance(arg, nodes.ObjectNode):
            parsed.append(arg.obj_id)
        elif isinstance(arg, nodes.DirectoryNode):
            parsed.append(arg.abspath)
        elif isinstance(arg, nodes.ResultsNode):
            parsed.extend(parse_args(*arg.children))
        elif isinstance(arg, dirs.PyosPath):
            parsed.append(arg)
        elif hist.is_obj_id(arg):
            parsed.append(arg)
        else:
            try:
                # Maybe it's an object id
                parsed.append(hist._ensure_obj_id(arg))
                continue
            except mincepy.NotFound:
                pass

            # Maybe it's a live object
            obj_id = hist.get_obj_id(arg)
            if obj_id is not None:
                parsed.append(obj_id)
            elif isinstance(arg, str):
                # Assume it's a path
                parsed.append(dirs.PyosPath(arg))
            else:
                raise TypeError("Unknown type '{}'".format(arg))

    return parsed



import typing

import mincepy

from .constants import NAME_KEY
from . import dirs
from . import nodes
from . import opts
from . import sopts
from . import queries

__all__ = 'init', 'reset'

from .dirs import path_to_meta_dict


def init():
    dirs.init()


def reset():
    dirs.reset()


def get_meta(*obj_ids: typing.Iterable) -> typing.List[dict]:
    """Get the metadata for a bunch of objects"""
    hist = mincepy.get_historian()
    meta = []
    for obj_id in obj_ids:
        meta.append(hist.meta.get(obj_id))
    return meta


def update_meta(*obj_ids: typing.Iterable, meta: dict):
    """Update the metadata for a bunch of objects"""
    hist = mincepy.get_historian()
    for obj_id in obj_ids:
        hist.meta.update(obj_id, meta)


def set_meta(*obj_ids: typing.Iterable, meta: dict):
    """Set the metadata for a bunch of objects"""
    hist = mincepy.get_historian()
    for obj_id in obj_ids:
        hist.meta.set(obj_id, meta)


def set_name(obj_id, name: str):
    """Set the name of the passed object(s)"""
    hist = mincepy.get_historian()
    update = {NAME_KEY: name}
    hist.meta.update(obj_id, update)


def get_name(*obj_or_ids) -> typing.Sequence[typing.Optional[str]]:
    """Get the name of the passed objects(s)"""
    hist = mincepy.get_historian()
    results = hist.meta.find({'obj_id': {'$in': obj_or_ids}})
    names = {meta['obj_id']: meta[NAME_KEY] for meta in results}
    return [names.get(obj_id, None) for obj_id in obj_or_ids]


def save(objects: typing.Sequence, paths: typing.Sequence[dirs.PathSpec] = None):
    if paths is None:
        metas = [None] * len(objects)
    else:
        assert len(objects) == len(paths), "Must supply equal number of objects and paths"
        metas = [path_to_meta_dict(path) for path in paths]

    if len(objects) == 1:
        metas = metas[0]

    return mincepy.get_historian().save(*objects, with_meta=metas)


def find(*args, **meta_filter) -> nodes.ResultsNode:
    options, starting_point = opts.separate_opts(*args)

    min_depth = options.pop(sopts.mindepth, 0)
    max_depth = options.pop(sopts.maxdepth, -1)

    if starting_point:
        spoints = [str(dirs.PyosPath(path).resolve()) for path in starting_point]
    else:
        spoints = [str(dirs.cwd())]

    meta_filter = meta_filter or {}

    meta_filter.update(
        queries.or_(*(queries.subdirs(point, min_depth, max_depth) for point in spoints)))

    hist = mincepy.get_historian()
    metas = hist.meta.find(meta_filter)

    results = nodes.ResultsNode()
    for meta in metas:
        results.append(nodes.ObjectNode(meta['obj_id']))

    return results
